@@ -10,7 +10,11 @@ from scipy.optimize import least_squares
 from formable.maths_utils import get_principal_values
 from formable.utils import requires, at_most_one_of
 from formable.yielding import (
-    YieldFunction, YIELD_FUNCTION_MAP, DEF_3D_RES, DEF_2D_RES)
+    YieldFunction,
+    get_yield_function_map,
+    DEF_3D_RES,
+    DEF_2D_RES,
+)
 from formable.yielding.yield_point import (
     YieldPointUnsatisfiedError, init_yield_point_criteria)
 
@@ -26,6 +30,8 @@ class LoadResponse(object):
     ALLOWED_DATA = [
         'true_stress',
         'equivalent_strain',
+        'equivalent_plastic_strain',
+        'accumulated_shear_strain',
     ]
 
     def __init__(self, **incremental_data):
@@ -53,14 +59,53 @@ class LoadResponse(object):
             msg = ('Pass at least one incremental data array. Allowed data is {}.')
             raise ValueError(msg.format(self.allowed_data_fmt))
 
+        outer_shape = incremental_data[list(incremental_data.keys())[0]].shape[0]
+        err = False
+        for i in incremental_data.values():
+            try:
+                if i.shape[0] != outer_shape:
+                    err = True
+            except AttributeError:
+                err = True
+            if err:
+                msg = ('All `incremental_data` dict values must be Numpy arrays of '
+                       'equal outer shape (corresponding to loading increments).')
+                raise ValueError(msg)
+
+        self._num_increments = outer_shape
+
         self._true_stress = incremental_data.pop('true_stress', None)
         self._equivalent_strain = incremental_data.pop('equivalent_strain', None)
+        self._equivalent_plastic_strain = incremental_data.pop(
+            'equivalent_plastic_strain', None)
+        self._accumulated_shear_strain = incremental_data.pop(
+            'accumulated_shear_strain', None
+        )
 
         if incremental_data:
             unknown_fmt = ', '.join(['"{}"'.format(i) for i in incremental_data])
             msg = ('Unknown incremental data "{}". Allowed incremental data are: {}')
             raise ValueError(msg.format(unknown_fmt, self.allowed_data_fmt))
 
+    @property
+    def num_increments(self):
+        return self._num_increments
+
+    def __len__(self):
+        return self.num_increments
+
+    def __repr__(self):
+        inc_data = []
+        for i in self.incremental_data_names:
+            inc_data.append(f'{i}{getattr(self, i).shape[1:] or (1,)}')
+        out = (
+            f'{self.__class__.__name__}('
+            f'num_increments={len(self)}, '
+            f'incremental_data={inc_data}'
+            f')'
+        )
+        return out
+
     @property
     def incremental_data_names(self):
         'Which of the allowed incremental data does this LoadResponse have?'
@@ -70,6 +115,14 @@ class LoadResponse(object):
                 has.append(allowed)
         return has
 
+    @property
+    def incremental_data(self):
+        'Get all incremental data as a dict.'
+        out = {}
+        for i in self.incremental_data_names:
+            out.update({i: getattr(self, i)})
+        return out
+
     @staticmethod
     def required_incremental_data(cls, analysis_type=None):
         """Get a list of the required incremental data for a given type of analysis."""
@@ -88,6 +141,14 @@ class LoadResponse(object):
     def equivalent_strain(self):
         return self._equivalent_strain
 
+    @property
+    def equivalent_plastic_strain(self):
+        return self._equivalent_plastic_strain
+
+    @property
+    def accumulated_shear_strain(self):
+        return self._accumulated_shear_strain
+
     @property
     @requires('true_stress')
     def principal_true_stress(self):
@@ -193,9 +254,32 @@ class LoadResponseSet(object):
         self.yield_stresses = []        # Appended in `self.calculate_yield_stresses`
         self.yield_functions = []       # Appended in `self.calculate_yield_function_fit`
 
+    def __repr__(self):
+        inc_data = []
+        for i in self.incremental_data_names:
+            inc_data.append(f'{i}{getattr(self.responses[0], i).shape[1:] or (1,)}')
+        out = (
+            f'{self.__class__.__name__}('
+            f'num_responses={len(self)}, '
+            f'incremental_data={inc_data}'
+            f')'
+        )
+        return out
+
     def __len__(self):
         return len(self.responses)
 
+    @property
+    def incremental_data_names(self):
+        """Which of the allowed LoadResponse incremental data do the constituent
+        LoadResponse objects have?"""
+        response = self.responses[0]
+        has = []
+        for allowed in response.ALLOWED_DATA:
+            if getattr(response, allowed) is not None:
+                has.append(allowed)
+        return has
+
     @init_yield_point_criteria
     def calculate_yield_stresses(self, yield_point_criteria):
         """Calculate and store yield stresses for each load case and the specified
@@ -218,6 +302,7 @@ class LoadResponseSet(object):
                 'response_idx': [],
             }
 
+            num_excluded = 0
             for resp_idx, resp_i in enumerate(self.responses):
 
                 val, _ = resp_i.get_yield_stress(yield_point_criteria, value_idx=ypv_idx)
@@ -225,9 +310,14 @@ class LoadResponseSet(object):
                 if val.size:
                     yield_stress_dict['values'].append(val)
                     yield_stress_dict['response_idx'].append(resp_idx)
+                else:
+                    num_excluded += 1
 
             yield_stress_dict['values'] = np.array(yield_stress_dict['values'])
 
+            print(f'{num_excluded}/{len(self)} responses skipped for yield point '
+                  f'{yield_point_val} from {yield_point_criteria}.')
+
             if not len(yield_stress_dict['values']):
                 msg = (f'No yield stresses were found for yield point criteria '
                        f'value {yield_point_val}.')
@@ -256,7 +346,7 @@ class LoadResponseSet(object):
                     if not isinstance(uniaxial_response, LoadResponse):
                         raise TypeError('`uniaxial_response` must be a LoadResponse')
 
-                    if not uniaxial_response.is_uniaxial(tol=5e-3):
+                    if not uniaxial_response.is_uniaxial(tol=8e-2):
                         msg = ('Specified `uniaxial_response` does not appear to be '
                                'uniaxial.')
                         raise ValueError(msg)
@@ -296,16 +386,16 @@ class LoadResponseSet(object):
         Parameters
         ----------
         yield_function : str or YieldFunction class
-            The yield function to fit. Available yield functions can be displayed using: 
-                `from formable import AVAILABLE_YIELD_FUNCTIONS`
-                `print(AVAILABLE_YIELD_FUNCTIONS)`
+            The yield function to fit. Available yield functions can be displayed using:
+                `from formable import get_available_yield_functions`
+                `print(get_available_yield_functions())`
         yield_point_criteria_idx : list of (list of int of length 2), optional
             If specified, fit the yield function to only the yield stresses calculated
             from the specified yield point criteria value. By default, None, in which case
             the yield function is fitted to all available yield stresses.
         initial_params : dict, optional
             Any initial guesses for the fitting parameters. Mutually exclusive with
-            additional keyword arguments passed, which are considered to be fixed.            
+            additional keyword arguments passed, which are considered to be fixed.
         kwargs : dict
             Any yield function parameters to be fixed during the fit can be passed as
             additional keyword arguments.
@@ -313,6 +403,7 @@ class LoadResponseSet(object):
         """
 
         # Validation:
+        YIELD_FUNCTION_MAP = get_yield_function_map()
         msg = 'Yield function "{}" not known. Available yield functions are: {}'
         available = ', '.join(['"{}"'.format(i) for i in YIELD_FUNCTION_MAP])
         msg = msg.format(yield_function, available)
@@ -387,7 +478,8 @@ class LoadResponseSet(object):
     def show_yield_functions_3D(self, normalise=True, resolution=DEF_3D_RES,
                                 equivalent_stress=None, min_stress=None, max_stress=None,
                                 show_axes=True, planes=None, backend='plotly',
-                                join_stress_states=False, show_contour_grid=False):
+                                show_stress_states=True, join_stress_states=False,
+                                show_contour_grid=False, layout=None,):
         'Visualise all fitted yield functions and data in 3D.'
 
         if not self.yield_functions:
@@ -405,6 +497,10 @@ class LoadResponseSet(object):
             yld_stresses.append(yld_stress_principal)
             stress_indices.append(resp_idx)
 
+        if not show_stress_states:
+            yld_stresses = None
+            stress_indices = None
+
         return YieldFunction.compare_3D(
             yld_funcs,
             normalise=normalise,
@@ -419,12 +515,15 @@ class LoadResponseSet(object):
             backend=backend,
             join_stress_states=join_stress_states,
             show_contour_grid=show_contour_grid,
+            layout=layout,
         )
 
     def show_yield_functions_2D(self, plane, normalise=True, resolution=DEF_2D_RES,
                                 equivalent_stress=None, min_stress=None, max_stress=None,
                                 show_axes=True, up=None, show_contour_grid=False,
-                                join_stress_states=False):
+                                show_stress_states=True, join_stress_states=False,
+                                show_numerical_lankford=False,
+                                show_numerical_lankford_fit=False, layout=None):
         'Visualise all fitted yield functions and data in 2D.'
 
         if not self.yield_functions:
@@ -442,6 +541,10 @@ class LoadResponseSet(object):
             yld_stresses.append(yld_stress_principal)
             stress_indices.append(resp_idx)
 
+        if not show_stress_states:
+            yld_stresses = None
+            stress_indices = None
+
         return YieldFunction.compare_2D(
             yld_funcs,
             plane,
@@ -455,4 +558,10 @@ class LoadResponseSet(object):
             up=up,
             show_contour_grid=show_contour_grid,
             join_stress_states=join_stress_states,
+            show_numerical_lankford=show_numerical_lankford,
+            show_numerical_lankford_fit=show_numerical_lankford_fit,
+            layout=layout,
         )
+
+    def show_yield_stresses_2D(self):
+        pass
