@@ -14,6 +14,7 @@ from plotly.colors import DEFAULT_PLOTLY_COLORS
 from scipy.optimize import least_squares, curve_fit
 
 from formable import utils, maths_utils
+from formable.yielding.map import get_yield_function_map
 
 DEF_3D_RES = 50
 DEF_2D_RES = 100
@@ -87,6 +88,19 @@ class YieldFunction(metaclass=abc.ABCMeta):
         )
         return value
 
+    @classmethod
+    def from_name(cls, name, **parameters):
+        'Get a specific yield function from the name and parameters.'
+        YIELD_FUNCTION_MAP = get_yield_function_map()
+        if name not in YIELD_FUNCTION_MAP:
+            available = ', '.join([f'{i!r}' for i in YIELD_FUNCTION_MAP])
+            msg = (f'Yield function "{name}" not known. Available yield functions '
+                   f'are: {available}.')
+            raise ValueError(msg)
+
+        yld_func_class = YIELD_FUNCTION_MAP[name]
+        return yld_func_class(**parameters)
+
     @classmethod
     def from_fit(cls, stress_states, initial_params=None, force_fit=False, **kwargs):
         """Fit the yield function to yield stress states.
@@ -224,7 +238,7 @@ class YieldFunction(metaclass=abc.ABCMeta):
                    equivalent_stress=None, min_stress=None, max_stress=None,
                    show_axes=True, planes=None, stress_states=None, backend='plotly',
                    stress_indices=None, join_stress_states=False, show_contour_grid=False,
-                   legend_text=None):
+                   legend_text=None, layout=None):
         """Visualise one or more yield functions in 3D.
 
         Parameters
@@ -316,39 +330,6 @@ class YieldFunction(metaclass=abc.ABCMeta):
 
             for idx, (yld_func, i) in enumerate(zip(yield_functions, values_all)):
 
-                # print(f'i.shape: {i.shape}')
-                # print(f'i: {i}')
-
-                # valid_idx = np.isfinite(i)
-
-                # print(f'valid_idx len: {valid_idx.sum()}')
-
-                # i_valid = i[valid_idx]
-                # stress_X_valid = stress_X[valid_idx]
-                # stress_Y_valid = stress_Y[valid_idx]
-                # stress_Z_valid = stress_Z[valid_idx]
-
-                # print(f'i.shape: {i_valid.shape}')
-                # print(f'i: {i}')
-
-                # print(f'max(i_valid): {max(i_valid)}, min(i_valid): {min(i_valid)}')
-
-                # print(f'i_valid[:100]:\n{i_valid[:100]!r}')
-                # print(f'stress_X_valid[:100]:\n{stress_X_valid[:100]!r}')
-                # print(f'stress_Y_valid[:100]:\n{stress_Y_valid[:100]!r}')
-                # print(f'stress_Z_valid[:100]:\n{stress_Z_valid[:100]!r}')
-
-                near_zero_idx = np.where(np.abs(i) < 1e-3)
-                #print(f'near_zero_idx: {near_zero_idx}')
-
-                near_zero_stress = np.array([
-                    stress_X[near_zero_idx],
-                    stress_Y[near_zero_idx],
-                    stress_Z[near_zero_idx],
-                ]).T
-
-                #print(f'near_zero_stress: {near_zero_stress.shape}\n{near_zero_stress!r}')
-
                 name = f'{idx + 1}. {yld_func.__class__.__name__}'
                 yield_point = getattr(yld_func, 'yield_point', None)
                 if legend_text:
@@ -365,7 +346,6 @@ class YieldFunction(metaclass=abc.ABCMeta):
                     'name': name,
                     'opacity': 0.5,
                     'showlegend': True,
-                    # ----
                     'isomin': 0,
                     'isomax': 0,
                     'surface_count': 1,
@@ -374,25 +354,8 @@ class YieldFunction(metaclass=abc.ABCMeta):
                         [1, DEFAULT_PLOTLY_COLORS[idx]],
                     ],
                     'showscale': False,
-                    # ----
-                    # 'surface_count': 4,
-                    # 'showscale': True,
                 })
 
-                # fig_data.append({
-                #     'type': 'scatter3d',
-                #     'mode': 'markers',
-                #     'x': stress_X,
-                #     'y': stress_Y,
-                #     'z': stress_Z,
-                #     'marker': {
-                #         'color': i,
-                #         'showscale': True,
-                #     },
-                #     'name': name,
-                #     'opacity': 0.5,
-                # })
-
             fig_annots = []
 
             if planes:
@@ -518,6 +481,7 @@ class YieldFunction(metaclass=abc.ABCMeta):
                         },
                         'annotations': fig_annots,
                     },
+                    **(layout or {}),
                 },
             )
 
@@ -527,7 +491,9 @@ class YieldFunction(metaclass=abc.ABCMeta):
     def compare_2D(cls, yield_functions, plane, normalise=True, resolution=DEF_2D_RES,
                    equivalent_stress=None, min_stress=None, max_stress=None,
                    stress_states=None, up=None, show_contour_grid=False,
-                   stress_indices=None, join_stress_states=False, legend_text=None):
+                   stress_indices=None, join_stress_states=False, legend_text=None,
+                   show_numerical_lankford=False, show_numerical_lankford_fit=False,
+                   layout=None):
         """Visualise multiple yield functions in 2D.
 
         Parameters
@@ -600,6 +566,90 @@ class YieldFunction(metaclass=abc.ABCMeta):
             if yield_point:
                 name += f' ({yield_point})'
 
+            if show_numerical_lankford:
+
+                num_lank = yld_func.get_numerical_lankford()
+                name += f' [R = {num_lank["lankford"]:.3f}]'
+                fit_centre = num_lank['fit_centre']
+
+                if normalise:
+                    grid_x = num_lank['grid_points'][0] - fit_centre
+                    grid_x = (grid_x / eq_stress)
+                    grid_x += (fit_centre / eq_stress)
+
+                    grid_y = num_lank['grid_points'][1] / eq_stress
+                else:
+                    grid_x = num_lank['grid_points'][0] / 1e6
+                    grid_y = num_lank['grid_points'][1] / 1e6
+
+                if show_numerical_lankford_fit:
+                    fig_data.append(
+                        {
+                            'type': 'scatter',
+                            'x': grid_x[num_lank['grid_point_fitted_idx']],
+                            'y': grid_y[num_lank['grid_point_fitted_idx']],
+                            'mode': 'markers',
+                            'marker': {
+                                'size': 4,
+                                'color': DEFAULT_PLOTLY_COLORS[idx],
+                            },
+                            'name': 'Tangent fit data',
+                            'legendgroup': name,
+                        }
+                    )
+
+                tangent_vec = num_lank['tangent_vector']
+                tangent_mag = max_stress * 0.2
+                t1 = - (0.5 * tangent_mag * tangent_vec)
+                t2 = + (0.5 * tangent_mag * tangent_vec)
+                tangent_line = np.vstack([t1, t2]).T
+
+                normal_vec = num_lank['normal_vector']
+                normal_vec *= tangent_mag * 1.3
+                normal_line_seg = np.array([
+                    [0, 0 + normal_vec[0]],
+                    [0, normal_vec[1]]
+                ])
+
+                if normalise:
+                    normal_line_seg[0] += (fit_centre / eq_stress)
+                    tangent_line[0] += (fit_centre / eq_stress)
+                else:
+                    normal_line_seg = normal_line_seg / 1e6
+                    normal_line_seg[0] += (fit_centre / 1e6)
+                    tangent_line = tangent_line / 1e6
+                    tangent_line[0] += (fit_centre / 1e6)
+
+                fig_data.append(
+                    {
+                        'type': 'scatter',
+                        'x': tangent_line[0],
+                        'y': tangent_line[1],
+                        'mode': 'lines',
+                        'line': {
+                            'color': DEFAULT_PLOTLY_COLORS[idx],
+                        },
+                        'showlegend': False,
+                        'name': name,
+                        'legendgroup': name,
+                    },
+                )
+
+                fig_data.append(
+                    {
+                        'type': 'scatter',
+                        'x': normal_line_seg[0],
+                        'y': normal_line_seg[1],
+                        'mode': 'lines',
+                        'line': {
+                            'color': DEFAULT_PLOTLY_COLORS[idx],
+                        },
+                        'showlegend': False,
+                        'name': name,
+                        'legendgroup': name,
+                    },
+                )
+
             fig_data.append({
                 'type': 'contour',
                 'x': grid_coords_2D[0],
@@ -621,6 +671,7 @@ class YieldFunction(metaclass=abc.ABCMeta):
                 'showlegend': True,
                 'hoverinfo': 'none',
                 'name': name,
+                'legendgroup': name,
             })
 
         if stress_states is not None:
@@ -836,7 +887,9 @@ class YieldFunction(metaclass=abc.ABCMeta):
                     'xanchor': 'center',
                     'y': -0.15,
                     'x': 0.5,
-                }
+                    'tracegroupgap': 5,
+                },
+                **(layout or {}),
             }
         )
 
@@ -1147,7 +1200,8 @@ class YieldFunction(metaclass=abc.ABCMeta):
     def show_3D(self, normalise=True, resolution=DEF_3D_RES, equivalent_stress=None,
                 min_stress=None, max_stress=None, show_axes=True, planes=None,
                 stress_states=None, backend='plotly', stress_indices=None,
-                join_stress_states=False, show_contour_grid=False, legend_text=None):
+                join_stress_states=False, show_contour_grid=False, legend_text=None,
+                layout=None):
         """
         Parameters
         ----------
@@ -1170,12 +1224,15 @@ class YieldFunction(metaclass=abc.ABCMeta):
             join_stress_states=join_stress_states,
             show_contour_grid=show_contour_grid,
             legend_text=legend_text,
+            layout=layout,
         )
 
     def show_2D(self, plane, normalise=True, resolution=DEF_2D_RES,
                 equivalent_stress=None, min_stress=None, max_stress=None,
                 stress_states=None, up=None, show_contour_grid=False, stress_indices=None,
-                join_stress_states=False, legend_text=None):
+                join_stress_states=False, legend_text=None,
+                show_numerical_lankford=False, show_numerical_lankford_fit=False,
+                layout=None):
         """
         Parameters
         ----------
@@ -1197,9 +1254,12 @@ class YieldFunction(metaclass=abc.ABCMeta):
             stress_indices=stress_indices,
             join_stress_states=join_stress_states,
             legend_text=legend_text,
+            show_numerical_lankford=show_numerical_lankford,
+            show_numerical_lankford_fit=show_numerical_lankford_fit,
+            layout=layout,
         )
 
-    def get_numerical_lankford(self, fit_domain=0.05, resolution=100, tol=2e-2):
+    def get_numerical_lankford(self, fit_domain=0.05, resolution=300, tol=1e-4):
         """Get the Lankford coefficient by numerically fitting the tangent to the yield
         function.
 
@@ -1227,10 +1287,22 @@ class YieldFunction(metaclass=abc.ABCMeta):
         else:
             equiv_stress = getattr(self, 'equivalent_stress')
 
+        # Identify a uniaxial stress state that is close to the yield surface (it will
+        # be close to the "equivalent stress", but for some reason they don't always
+        # precisely coincide):
+        fit_centre_range = np.linspace(-0.5, 0.5, num=10000) * equiv_stress
+        fit_centre_range += equiv_stress
+        num_ss = fit_centre_range.shape[0]
+        stress_states = np.tile(np.eye(3), num_ss).reshape(num_ss, 3, 3)
+        stress_states[:, 0, 0] = fit_centre_range
+        fit_centre_vals = self.get_value(stress_states)
+        fit_centre_idx = np.argmin(np.abs(fit_centre_vals))
+        fit_centre = fit_centre_range[fit_centre_idx]
+
         X, Y = np.meshgrid(*[np.linspace(-0.5, 0.5, num=resolution)] * 2)
         xy = np.vstack([X.flatten(), Y.flatten()])
-        xy *= equiv_stress * fit_domain
-        xy += np.array([[equiv_stress, 0]]).T
+        xy *= fit_centre * fit_domain
+        xy += np.array([[fit_centre, 0]]).T
 
         num_ss = xy.shape[1]
         stress_states = np.tile(np.eye(3), (num_ss, 1)).reshape(num_ss, 3, 3)
@@ -1238,39 +1310,29 @@ class YieldFunction(metaclass=abc.ABCMeta):
         stress_states[:, 1, 1] = xy[1]
 
         vals = self.get_value(stress_states)
-
-        # print(f'vals:\n{vals}\n')
         good_idx = np.where((np.abs(vals) - tol) < 0)[0]
         if not good_idx.size:
             raise ValueError(f'No yield surface solutions within `tol={tol}`.')
 
-        # print(f'good_idx:\n{good_idx}\n')
-
-        # [# For fitting, normalise x, y values by equivalent stress (we just want the
-        # # direction of the tangent vector in the σ_3 = 0 plane)] ?
-        # xy_norm = xy / equiv_stress
-
         fit_coords = xy[:, good_idx]
-        # Get x, y values to fit:
-
-        # print(f'fit_coords:\n{fit_coords}\n')
-
         popt, pcov = curve_fit(maths_utils.line, fit_coords[0], fit_coords[1])
-        # print(f'popt: {popt}\n; pcov:\n{pcov}\n')
+        m = popt[0]
 
-        m, c = popt
-        if c < 0:
-            m, c = -m, -c
-
-        tangent_x = [np.min(xy[0]), np.max(xy[0])]
-        tangent_y = maths_utils.line(np.array(tangent_x), m, c)
-
-        normal = np.array([-m, 1])
-        # print(f'normal: {normal}')
+        tangent_vector = np.array([1, m])
+        tangent_vector_unit = tangent_vector / np.linalg.norm(tangent_vector)
 
+        normal = -1 * np.array([-m, 1])  # -1 to point out from the surface
         normal_n = normal / np.linalg.norm(normal)
-        # print(f'normal_n: {normal_n}')
 
-        lankford = normal_n[1] / (-normal[0] - normal[1])
+        lankford = normal_n[1] / -(normal_n[0] + normal_n[1])
 
-        return (lankford, (tangent_x, tangent_y))
+        out = {
+            'grid_points': xy,
+            'grid_point_fitted_idx': good_idx,
+            'normal_vector': normal_n,
+            'tangent_vector': tangent_vector_unit,
+            'lankford': lankford,
+            'fit_centre': fit_centre,
+        }
+
+        return out
